class ChainedMethods:

    def __init__(self, name=None, parent=None):
        self._parent = parent
        self._name = name if name else type(self).__name__
        self._args = tuple()
        self._kwargs = dict()
        self._child = None

    def _root(self):
        if self._parent:
            return self._parent._root()
        else:
            return self

    def __getattr__(self, name):
        self._child = child = type(self)(name, self)
        return child

    def __call__(self, *args, **kwargs):
        self._kwargs = kwargs
        self._args = args
        return self

    def __repr__(self):
        # rework using self._descriptor()
        if self._parent:
            parent = repr(self._parent)
            return '{}.{}{}'.format(
                parent,
                self._name, str((self._args, self._kwargs))
            )
        else:
            return self._name

    def _descriptor(self):
        if self._parent:
            yield from self._parent._descriptor()
            yield (self._name, self._args, self._kwargs)
        else:
            yield self._name

    def serialize(self):
        return list(self._descriptor())


# XXX:
#
# x = ChainedMethods('graph').vertex(label="foo").outgoings(label="subscribe")
# rpr = 'outgoings'.'vertex'.graph(((), {'label': 'foo'}))(((), {'label': 'subscribe'}))
