from bsddb3.db import DB
from bsddb3.db import DBEnv
from bsddb3.db import DB_HASH
from bsddb3.db import DB_CREATE
from bsddb3.db import DB_INIT_LOG
from bsddb3.db import DB_INIT_TXN
from bsddb3.db import DB_INIT_MPOOL
from bsddb3.db import DB_MULTIVERSION
from bsddb3.db import DB_RECOVER_FATAL
from bsddb3.db import *

from .chained_methods import ChainedMethods


class BrokenIndexConstraint(Exception):
    pass


class Index:

    def __init__(self, descriptor, storage):
        self._descriptor = descriptor
        self._storage = storage

    def label(self, element, expected):
        if element.label() == expected:
            return ''
        else:
            return None

    def _update(self, element):
        keys = list()
        unique = False
        for name, args in self._descriptor:
            if name == 'unique':
                unique = True
                # unique is only supported at the end
                # of an index descriptor
                break
            else:
                method = getattr(self, name, None)
                if method:
                    key = method(element, *args)
                    if key is None:
                        return
                    elif key:
                        keys.append(key)
                    # else:
                    #     element match but doesn't provide a key
                else:
                    try:
                        key = element[name]
                    except:
                        return
                    try:
                        old = element.old(name)
                    except KeyError:
                        old = None
                    finally:
                        keys.append((key, old))
        if keys:
            value = element.identifier()
            key = ':'.join(map(lambda x: x[0], keys))
            try:
                old = ':'.join(map(lambda x: x[1], keys))
            except TypeError:
                # it's a new index entry
                self._storage.put(key, value)
            else:
                # remove old entry and add new one
                self._storage.delete(old)
                self._storage.put(key, value)


class IndexDescriptor(ChainedMethods):

    def __init__(self, txn, name=None, parent=None):
        self._txn = txn
        super().__init__(name, parent)

    def _new(self, parent, name):
        return type(self)(self._txn, name, self)

    def __iter__(self):
        next = self._child
        while next:
            yield next._name, next._args
            next = next._child

    def _save(self):
        # XXX: need also to take care of unique/not-unique indexes
        # not unique indexes can take several values.
        # maybe it will be better to create a database for those
        # maybe a sequence
        ordered = False
        for operator, args in descriptor:
            if operator == 'order_by':
                ordered = True
                break
        txn = self._txn._txn
        env = self._txn._database.env
        self.storage = storage = DB(env)
        root = self._root()
        index = Index(root, storage)
        self._txn._indexes.add(index)
        # open database
        flags = DB_CREATE | DB_MULTIVERSION
        name = 'ajgu-index:%s' % name
        backend = DB_BTREE if ordered else DB_HASH
        storage.open(name, None, backend, flags, 0, txn=txn)        
