from .chained_methods import ChainedMethods


class StopTraversal(Exception):
    pass


class QueryProcessor:

    def __init__(self, txn, query, elements):
        self.txn = txn
        self.query = query
        self.elements = elements

    def __iter__(self):
        if self.query:
            query = self.query
            next_operator, query_queue = query[0], query[1:]
            method_name, args = next_operator
            method = getattr(self, method_name)
            for element in self.elements:
                yield from method(query_queue, element, *args)
        else:
            yield from self.elements

    def label(self, query, element, label):
        if element.label() == label:
            yield from QueryProcessor(self.txn, query, (element,))

    def outgoings(self, query, element):
        for element in element.outgoings():
            yield from QueryProcessor(self.txn, query, (element,))

    def incomings(self, query, element):
        for element in element.incmoings():
            yield from QueryProcessor(self.txn, query, (element,))

    def filter(self, query, element, specification):
        property, operator, value = specification
        if operator != 'eq':
            raise NotImplementedError
        try:
            if element[property] == value:
                yield from QueryProcessor(self.txn, query, (element,))
        except KeyError:
            pass

    def one(self, query, element, test):
        yield element
        raise StopTraversal()

    def end(self, query, element):
        yield from QueryProcessor(self.txn, query, (element.end(),))
        

class Query(ChainedMethods):
    
    def __init__(self, txn, name=None, parent=None):
        self._txn = txn
        super().__init__(name, parent)

    def _new(self, parent, name):
        return type(self)(self._txn, name, self)

    def one(self):
        return list(self)[0]

    def all(self):
        return list(self)

    def __iter__(self):
        root = self._root()
        _, descriptor = self._descriptor()
        yield from QueryProcessor(self._txn, descriptor, self._txn.all())
